/*
 * mp: educational resources to teach concurrent programming
 * Copyright (C) 2016  Renzo Davoli, University of Bologna
 *      
 * implementation message passing services
 *        
 * sm is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *             
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; If not, see <http://www.gnu.org/licenses/>.
 *                   
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <suspend.h>
#include <pthread.h>
#include <signal.h>
#include <errno.h>
#include <msg.h>
#include <ctype.h>

#define NONE 0
#define ASYNCHRONOUS 'A'
#define COMPLETELY_ASYNCHRONOUS 'N'
#define SYNCHRONOUS 'S'

#define mutex_in(X) pthread_mutex_lock(X)
#define mutex_out(X) pthread_mutex_unlock(X)

#define MAXLEN 4096

static int fdsocket = -1;
static pthread_t msgh_t;
static pthread_t app_t;
static char *w4sender;
static int mp_model;

struct msgq {
	struct msgq *next;
	size_t payloadlen;
	char type;
	char payload[0];
};

static pthread_mutex_t mgsq_mutex = PTHREAD_MUTEX_INITIALIZER;
static struct msgq *msgqhead, *msgqtail;

static inline int oktype(int sendtype, int recvtype) {
	return sendtype == recvtype || sendtype == tolower(recvtype);
}

/* add an incoming message to the buffer,
	 wakeup the app thread if it is waiting for a message from this sender
 */
static void msgqadd(char *sender, const void *buf, size_t len, char type) {
	size_t newlen = sizeof(struct msgq) + len + strlen(sender) + 1;
	struct msgq *new = malloc(newlen);
	if (new) {
		new->next = NULL;
		new->payloadlen = len;
		new->type = type;
		memcpy(new->payload, buf, len);
		strcpy(new->payload + len, sender);
		mutex_in(&mgsq_mutex);
		if (msgqhead == NULL)
			msgqhead = new;
		else
			msgqtail->next = new;
		msgqtail = new;
		if (w4sender != NULL &&
				(*w4sender == 0 || strcmp(w4sender, sender) == 0)) {
			w4sender = NULL;
			wakeup(app_t);
		} else
			mutex_out(&mgsq_mutex);
	}
}

/* get a message from the buffer.
	 suspend the app thread if blocking and no matching msg is in the buffer 
 */
size_t msgget(char *sender, void *buf, size_t len, int blocking, char type) {
	struct msgq **scan;
	size_t pktlen;
	if (buf == NULL)
		type = tolower(type);
	mutex_in(&mgsq_mutex);
	do {
		for (scan = &msgqhead; *scan != NULL; scan = &((*scan)->next)) {
			if (oktype((*scan)->type,type) && 
					(sender == NULL || *sender == 0 || 
					 strcmp(sender, (*scan)->payload + (*scan)->payloadlen) == 0))
				break;
		}
		if (*scan == NULL && blocking) {
			w4sender = sender ? sender : "";
			mutex_out(&mgsq_mutex);
			suspend();
		}
	} while (blocking && *scan == NULL);
	if (*scan) {
		struct msgq *this = *scan;
		*scan = this->next;
		pktlen = this->payloadlen;
		if (pktlen < len)
			len = pktlen;
		memcpy(buf, this->payload, len);
		if (sender && *sender == 0)
			strncpy(sender, this->payload + pktlen, MSGNAMELEN);
		mutex_out(&mgsq_mutex);
		free(this);
	} else {
		mutex_out(&mgsq_mutex);
		pktlen = -1;
	}
	return pktlen;
}

/* common code for sending a message */
static ssize_t msg_asend(char *dest, const void *buf, size_t len, char type) {
	struct iovec iov[]={{&type,1},{(void *)buf, len}};
	struct sockaddr_un sk;
	struct msghdr hdr={&sk, sizeof(sk), iov, sizeof(iov)/sizeof(iov[0]), NULL, 0, 0};
	ssize_t rval;
	if (buf == NULL)
		type = tolower(type);
	sk.sun_family = AF_UNIX;
	strncpy(sk.sun_path, dest, 108);
	rval = sendmsg(fdsocket, &hdr, 0);
	return (rval > 0) ? rval - 1 : rval;
}

ssize_t msg_send(char *dest, const void *buf, size_t len) {
	switch (mp_model) {
		case ASYNCHRONOUS:
		case COMPLETELY_ASYNCHRONOUS:
			return msg_asend(dest, buf, len, mp_model);
		case SYNCHRONOUS:
			{
				ssize_t rval=msg_asend(dest, buf, len, mp_model);
				if (rval >= 0)
					msgget(dest, NULL, 0, 1, mp_model);
				return rval;
			}
		default:
			errno = EPROTO;
			return -1;
	}
}

ssize_t msg_recv(char *sender, void *buf, size_t len) {
	switch (mp_model) {
		case ASYNCHRONOUS:
			return msgget(sender, buf, len, 1, mp_model);
		case COMPLETELY_ASYNCHRONOUS:
			return msgget(sender, buf, len, 0, mp_model);
		case SYNCHRONOUS:
			{
				ssize_t rval;
				if (sender == NULL) {
					msg_procname sender = "";
					rval = msgget(sender, buf, len, 1, mp_model);
					if (rval >= 0)
						msg_asend(sender, NULL, 0, mp_model);
				} else {
					rval = msgget(sender, buf, len, 1, mp_model);
					if (rval >= 0)
						msg_asend(sender, NULL, 0, mp_model);
				}
				return rval;
			}
		default:
			return -1;
	}
}

static void *msgh(void *arg) {
	ssize_t len;
	char buf[MAXLEN];
	char type;
	struct iovec iov[]={{&type,1},{(void *)buf, MAXLEN}};
	struct sockaddr_un sk;
	struct msghdr hdr={&sk, sizeof(sk), iov, sizeof(iov)/sizeof(iov[0]), NULL, 0, 0};
	while ((len = recvmsg(fdsocket, &hdr, 0)) > 0) {
		msgqadd(sk.sun_path, buf, len - 1, type);
		hdr.msg_namelen = sizeof(sk);
	}
	return NULL;
}

static void common_msg_init(char *name) {
	struct sockaddr_un sk;
	fdsocket = socket(AF_UNIX, SOCK_DGRAM, 0);
	sk.sun_family = AF_UNIX;
	strncpy(sk.sun_path, name, 108);

	if (bind(fdsocket, (struct sockaddr *) &sk, sizeof(sk)) < 0) {
		perror("bind");
		exit(1);
	}
	app_t = pthread_self();
	pthread_create(&msgh_t, NULL, msgh, name);
}

void msg_myname(char *myname) {
	struct sockaddr_un sk;
	socklen_t sklen = sizeof(sk);
	if (fdsocket >= 0 &&
			getsockname(fdsocket, (struct sockaddr *) &sk, &sklen) == 0) 
		strcpy(myname,sk.sun_path);
	else
		*myname = 0;
}

void async_msg_init(char *name) {
	common_msg_init(name);
	mp_model = ASYNCHRONOUS;
}

void sync_msg_init(char *name) {
	common_msg_init(name);
	mp_model = SYNCHRONOUS;
}

void nb_msg_init(char *name) {
	common_msg_init(name);
	mp_model = COMPLETELY_ASYNCHRONOUS;
}

void msg_fini(void) {
	struct sockaddr_un sk;
	socklen_t sklen = sizeof(sk);
	if (fdsocket >= 0 &&
			getsockname(fdsocket, (struct sockaddr *) &sk, &sklen) == 0) {
		unlink(sk.sun_path);
		close(fdsocket);
		mp_model = NONE;
		fdsocket = -1;
	}
}

static void sig_handler(int sig)
{
	fprintf(stderr,"Caught signal %d, cleaning up and exiting\n", sig);
	msg_fini();
	signal(sig, SIG_DFL);
	if (sig == SIGTERM)
		_exit(0);
	else
		kill(getpid(), sig);
}

__attribute__((constructor)) void _msg_init() {
	int signals[] = {SIGHUP, SIGINT, SIGTERM, SIGQUIT, SIGILL, SIGABRT, SIGFPE, SIGSEGV, SIGPIPE, SIGBUS};
	int i;
	for (i = 0; i < sizeof(signals) / sizeof(signals[0]); i++)
		signal(signals[i], sig_handler);
}

__attribute__((destructor)) void _msg_fini() {
	msg_fini();
}
